function results = dclde2018score(gt, detections, varargin)
% results = dclde2015score(gt, detections, OptArgs)
%
% Compute precision/recall score and other statistics for sets of files
% related to a single location.  To compute statistics for multiple
% sites, use dclde2018multiscore.
%
% gt - ground truthed detections
% detections - detections to be evaluated
% gt and detections should either be:
%   filenames - Assumes that these are DCLDE 2018 format CSV files
%       and reads the files via read_annotations
%   structures - Structure format must be the same as that returned
%       by read_annotations (see read_annotations for details)
%
% Optional arguments
%     'IgnoreLabels', true|false (default false)
%        if true, only detection times are considered.  
%        Species and call names are ignored.  This can be useful for
%        testing signal detection algorithms.
%     'RemoveZeroDuration', true|false (default true)
%         Remove detections that have a duration of 0 s
%     'SpecificLabels', mapObj
%        Only consider detections from specified species (and possibly
%        specific calls from that species) when scoring.  All other
%        detections are not scored.  See read_annotations for details.
%     'AllQualities', true|false (default false)'
%        If false, only score on "good" quality calls. If true, consider 
%        all calls regardless of call quality. 
%
% CAVEATS:
%   Optional arguments RemoveZeroDuration and SpecificLabels are
%   only processed when gt and/or detections contain CSV filenames as
%   this functionality is implemented in the CSV reader function
%   read_annotations.
%
%   Degenerate cases of precision and recall return 1 which can be
%   misleading.  Examples:
%     Nothing to detect:  Recall = 1
%     Nothing detected:  Precision = 1
%
% See also: read_annotations
%
% Author:  Marie A. Roch, San Diego State University

% a bit of basic error checking...
narginchk(2, Inf);
if ischar(gt)
    gt = {gt};
end
if ischar(detections)
    detections = {detections};
end
if (~ isstruct(gt) && ~ iscell(gt)) || ...
   (~ isstruct(detections) && ~ iscell(detections))
    error('dclde:arguments', ...
        'Ground truth and detections must be strings, structures, or cell arrays.');
end

vidx = 1;
IgnoreLabels = false;  % default behavior
AllQualities = false; % default behavior
ScoreLabels = [];
while vidx <= length(varargin)
    switch varargin{vidx}
        case 'IgnoreLabels'
            IgnoreLabels = varargin{vidx+1};
            varargin(vidx:vidx+1) = [];  % Remove processed arg
        case 'AllQualities'
            AllQualities = varargin{vidx+1};
            varargin(vidx:vidx+1) = [];  % Remove processed arg
        otherwise
            vidx = vidx+2;  % skip argument
    end
end

% ground-truth labeled detections
if isstruct(gt)
    truth = gt;
else
    truth = read_annotations(gt, varargin{:});
end
% detections to be scored
if ~ isstruct(detections)
    detections = read_annotations(detections, varargin{:});
end

% Various bookkeeping structures -----------------------

% Number of detections matched to each ground truth detection
% Used to track fragmented detections
truthMatches = zeros(truth.N, 1);
% Amount of time of each ground truth detection that was matched.
% When divived by ground truth duration, it gives coverage - 
% the percent of a ground truth that was detected.
% 
% Low truth coverage indicates that we are missing parts of detections
% covered by the ground truth.
truthCoverage = zeros(truth.N, 1);
% Amount of time that a detection could be matched to ground truth
% When devided by the detection duration, it gives a different
% kind of coverage, percentage of time that a detection could be matched
% to ground truth.
%
% Low detection coverage indicates that we are over detecting relative 
% to the ground truth reports. (Related to false positives)
detectionCoverage = zeros(detections.N, 1);

goodMatches = ones(truth.N, 1);
% False positive indicator function.  
% Assume false positive until we find a match.
falseposI = ones(detections.N, 1);

% --------------------------------------------------------------------
% Algorithm roughly follows the idea of a line scan algorithm.
% We scan the ground truth detections linearly, keeping track of the
% first possible candidate match amongst the detections and examining
% from there until we hit a detection that could not possibly match
% the current ground truth detection.

didx = 1;  % First possible matching detection
% Predicate to note whether call field is present
callP.detections = isfield(detections, 'Call');
callP.truth = isfield(truth, 'Call');
qualityP.truth = isfield(truth, 'quality');
 
% associations(gtidx, detidx) - True if gtidx'th ground truth
% detection overlaps with the detidx'th detection
associations = sparse(truth.N, detections.N);
ignoreDetection = sparse(truth.N, detections.N);
ignoreTruth = zeros(truth.N,1);
truth.NIgnore = 0;
if qualityP.truth && ~AllQualities
    % hold on to info about call quality
    ignoreTruth = ~strcmpi('good',truth.quality);
    truth.NIgnore = sum(ignoreTruth);
end

for gidx = 1:truth.N
    % If the first possible detection ends before the start
    % advance to the next one
    while didx <= detections.N && detections.end(didx) < truth.start(gidx)
        detectionCandidates(didx) = false;
        didx = didx + 1;
    end
    
    if didx <= detections.N
        % check each candidate up to the point that we have
        % a start that is after then end of the current detection
        current = didx;
        done = false;
        while ~ done
            % check for overlap
            if detections.start(current) <= truth.end(gidx) && ...
                detections.end(current) >= truth.start(gidx)
                % overlap occurred 
                
                % Make sure correct species and call type (if present)
                correct = IgnoreLabels || ... 
                    strcmp(detections.species{current}, truth.species{gidx});
                if callP.truth
                    % Ground truth expects call type, do they match?
                    correct = IgnoreLabels || ...
                        (correct && callP.detections && ...
                        strcmp(detections.call{current}, truth.call{gidx}));
                end

                % If it's not a "good" call, then flag it
                if ignoreTruth(gidx)
                    ignoreDetection(gidx, current) = ignoreTruth(gidx);
                end
                
                if correct
                    associations(gidx, current) = true;  % note match
                end
                
            end
            current = current + 1;
            % All done if no more detections or the next detection
            % starts after the current ground truth index
            done = current > detections.N || ...
                detections.start(current) > truth.end(gidx);
        end
        
        % Bookkeeping to track quality of a good detection.
        matches = find(associations(gidx,:));
        
        if ~ isempty(matches)
            % Matched one or more detections to this ground
            % truth detection
            % Note that the detections are not false positives
            falseposI(matches) = false;
            
            % Note how many detections matched this ground truth detections
            truthMatches(gidx) = length(matches);
%             if qualityP.truth %
%                 % note wether any of the matched detections were "good"
%                 % quality
%                 ignoreMatch(gidx) = min(ignoreDetection(gidx,matches));
%             end
            % Find amount of overlap between detections and this 
            % ground truth
            truthCoverage(gidx) = ...
                overlap(truth.start(gidx), truth.end(gidx), ...
                [detections.start(matches), detections.end(matches)]);  
        end
    end
end

% We have processed all of the ground truth detections.
% Next we look at the detections to determine what portion of each
% detection matched a ground truth item.
keepDetection = ones(detections.N,1);
for didx = 1:detections.N
    % Find ground truth items associated with this detection.
    matches = find(associations(:,didx));
    if ~isempty(matches)     
        keepDetection(didx) = ~min(ignoreDetection(matches,didx));

        detectionCoverage(didx) = ...
            overlap(detections.start(didx), detections.end(didx), ...
            [truth.start(matches), truth.end(matches)]);
    end
end
keepDetection = logical(keepDetection);
nIgnoreDetection = sum(~keepDetection);
% --------------------------------------------------------------------
% We have everything we need now, start computing results.

% Precision:  Ratio of good detections to detections
if detections.N > 0
    results.precision = ((detections.N-nIgnoreDetection) - sum(falseposI))...
        / (detections.N - nIgnoreDetection);
else
    % Degenerate case - no detections, 100% of detections are correct...
    results.precision = 1;
end
    
% Recall:  Ratio of ground truth items detected to # expected
% Don't penalize for ignored calls
if truth.N > 0 
    results.recall = sum(truthMatches(~ignoreTruth) > 0)/...
        (truth.N - truth.NIgnore);
else
    % Degenerate case - we detected everything there was to detect!
    results.recall = 1;
end

results.falsePosI = falseposI(keepDetection);  % false positive map for detections

% Show percentage of each ground truth that was covered as well
% as the overall coverage.
if truth.N
    results.truthCoverage = truthCoverage(~ignoreTruth) ;
    results.truthCoveragePct = truthCoverage(~ignoreTruth)  ./...
        truth.durations(~ignoreTruth);
    results.truthCoverageOverallPct = sum(truthCoverage(~ignoreTruth)) /...
        sum(truth.durations(~ignoreTruth));
else
    % No ground truth signals
    results.truthCoverage = NaN;
    results.truthCoveragePct = NaN;
    results.truthCoverageOverallPct = NaN;
end


% Show percentage of detections that covered a ground truth tonal
results.detectionCoverage = detectionCoverage(keepDetection);
results.detectionCoveragePct = detectionCoverage(keepDetection) ./ detections.durations(keepDetection);
results.detectionCoverageOverallPct = ...
    sum(detectionCoverage(keepDetection)) / sum(detections.durations(keepDetection));

% fragmentation shows the number of detections that covered a single
% ground truth detection
results.fragmentation = full(sum(associations(:,keepDetection),2));
% mean fragmentation is computed from detections
detectionI = results.fragmentation > 0;
results.Efragmentation = mean(results.fragmentation(detectionI));

results.associations = associations;
results.truth = truth;
results.detections = detections;
results.ignoreDetections = ~keepDetection;
results.ignoreTruth = ignoreTruth;
