function mergeddata = read_annotations(filenames, varargin)
% data = read_annotations(filenames, OptionalArgs)
% Return annotations from one or more CSV files.
% Multiple files must be in a cell array and are merged.
%
% Optional arguments
%     'RemoveZeroDuration', true|false (default true)
%         Remove detections that have a duration of 0 s
%     'SpecificLabels', mapObj
%        Only consider detections from specified species (and possibly
%        specific calls for that species) when scoring.  All other
%        detections are not scored.
%        mapObj must be an instance of containers.Map.class
%
%        map keys are species labels - only species found in the keys
%           will be scored
%        The value of each map key should be:
%           {} - empty cell array: no restriction on call type
%           {'call1', 'call2', ..., 'callN'} - a set of calls for
%                this species
%        Example:
%        map = containers.Map({'Bm','Bp'} {{'B', 'D'}, {}})
%        Score only B or D calls for blue whales and 
%        any call for fin whales.
%
%        See Matlab documentation of containers.Map.class for further
%        details
% data contains the following fields:
%
% N = number of annotations
% start - start times as Matlab serial dates (datenum)
% end - end times (serial dates)
% durations - derived duration of each annotation (serial dates)
% site - cell array of site names
% species - cell array of species names
% call - cell array of call names (if present in CSV)

narginchk(1,inf);

if ischar(filenames)
    filenames = {filenames};  % Ensure filenames is a cell array
end

vidx = 1;
RetainOnly = [];   % defaults
RemoveZeroDuration = false;
while vidx <= length(varargin)
    switch varargin{vidx}
        case 'RemoveZeroDuration'
            RemoveZeroDuration = varargin{vidx+1};
            vidx = vidx + 2;
        case 'SpecificLabels'
            RetainOnly = varargin{vidx+1};
            vidx = vidx + 2;
            okay = isobject(RetainOnly);
            if okay
                okay = strcmp(class(RetainOnly), 'containers.Map');
            end
            if ~ okay
                error('SpecificLabels must be a containers.Map');
            end
        otherwise
            error('Bad optional keyword/value pair');
    end
end
% expected fields in file
fields = {'project', 'site', 'species', 'start', 'end', 'quality'};
% expected fields to be populated
popfields = fields;
popfields{end+1} = 'durations';

for fidx = 1:length(filenames)
    filename = filenames{fidx};
    
    % Read the data
    h = fopen(filename, 'r');
    if h == -1
        error('DCLDE:FILE', 'Unable to open %s', filename);
    end
    csvdata = textscan(h, '%s %s %s %s %s %s', 'delimiter', ',');
    fclose(h);
    % Always reads in six fields, but quality is not always present
    if sum(strcmp(csvdata{end}, '')) == length(csvdata{end})
        csvdata(end) = [];  % Remove quality if none was given
    end

    % these fields should be converted to timestamps
    datefields = {'start', 'end'};
    for idx = 1:length(csvdata)
        % Is this a date field?
        if sum(strcmp(fields{idx}, datefields))
            data(fidx).(fields{idx}) = dbISO8601toSerialDate(csvdata{idx});
        else
            data(fidx).(fields{idx}) = csvdata{idx};
        end
    end
    
    bad = find(data(fidx).end - data(fidx).start < 0);
    if ~ isempty(bad)
        error('DCLDE:BADENTRY - "%s" end before start at entries:\n\t%s', ...
            filename, sprintf('%d ', bad));
    end
    
    if ~ issorted(data(fidx).start)
        warning('DLCDE:SORTORDER - "%s" not sorted by start - sorting...', ...
            filename);
        data(fidx) = sortbystart(data(fidx)); 
    end
    
    data(fidx).durations = data(fidx).end - data(fidx).start;

    % Process detections that have zero length
    zerolen = find(data(fidx).durations == 0);
    if ~ isempty(zerolen)
        % Either remove them or warn the user.
        if RemoveZeroDuration
            for fieldidx = 1:length(popfields)
                if isfield(data(fidx), popfields{fieldidx})
                    data(fidx).(popfields{fieldidx})(zerolen) = [];
                end
            end
        else
            %warning('File %s contains zero length durations\n%s\n', ...
            %    filename, fmtdetections(data(fidx), zerolen));
        end
    end
    
    % Warn user for really long durations
    toolongdays = 3;
    long = find(data(fidx).durations > toolongdays);
    if ~isempty(long)
        warning('File %s contains detections over %d days\n%s\n', ...
                filename,toolongdays, fmtdetections(data(fidx), long));
    end
end


% merge all data  ----------------------------------------
if length(filenames) > 1
    % Verify all files either have or don't have call parameter
    if isfield(data, 'quality')
        callcount = cellfun(@(x) length(x), {data.call});
        if sum(diff(callcount > 0))
            error('DCLDE:COMBINE', ...
                'Files must consistently have/omit quality field');
        end
    end
    % Concatenate fields of each file
    names = fieldnames(data);
    for idx=1:length(names)
        mergeddata.(names{idx}) = vertcat(data.(names{idx}));
    end
    
    % Need to sort
    mergeddata = sortbystart(mergeddata);
else
    mergeddata = data(1);  % nothing to merge
end
mergeddata.N = length(mergeddata.start);

mergeddata = removeduplicates(mergeddata, filenames);

% Remove any undesired data
if ~isempty(RetainOnly)
    % User wants only certain species and calls retained
    mergeddata = filterdata(mergeddata, RetainOnly);
end
     
        
    
function data = sortbystart(data)

% sort by start date
[data.start, permutations] = sort(data.start);

fields = fieldnames(data);
for fidx = 1:length(fields)
    if ~ (strcmp(fields{fidx}, 'start') || ...
          isscalar(data.(fields{fidx})) || ...
          isempty(data.(fields{fidx})))
        data.(fields{fidx}) = data.(fields{fidx})(permutations);
    end
end
    
function data = removeduplicates(data, filenames)
% data = removeduplicates(data)
% Given a set of CSV annotation, remove any duplicates

fields = fieldnames(data);

% Verify no duplicate entries
if data.N > 1
    % Guaranteed sorted, check for consecutive identical start/end
    candidates = find(diff(data.start) == 0 & ...
        diff(data.end) == 0);
    
    if ~isempty(candidates)
        % start and ends match, check for other parameters
        for fidx = 1:length(fields)
            switch fields{fidx}
                case {'start', 'end', 'N', 'durations'}
                    % do nothing
                otherwise
                    % Expecting everything else to be strings
                    identical = strcmp(...
                        data.(fields{fidx})(candidates), ...
                        data.(fields{fidx})(candidates+1));
                    % Get rid of anything that's different
                    candidates = candidates(identical);
                    if isempty(candidates)
                        break
                    end
            end
        end
        
        % Remove any duplicates
        if ~ isempty(candidates)
            warning('dclde2015:duplicate_entries', ...
                'duplicate entries lines %s\nFile set:\n%s\nFirst line of each duplicate\n%s', ...
                sprintf('%d ', candidates), ...
                sprintf('\t%s\n', filenames{:}), ...
                fmtdetections(data, candidates));
            
            for fidx = 1:length(fields)
                if ~ strcmp(fields{fidx}, 'N')
                    data.(fields{fidx})(candidates) = [];
                end
            end
            data.N = length(data.start);
        end
    end
end

function data = filterdata(data, map)
% data = filterdata(data, map)

retain = false(size(data.start));
species = map.keys();  % Retrieve the species to be filtered.
callwarn = false;
callP = isfield(data, 'call');  % call present predicate

for sidx = 1:length(species)
    smatch = strcmp(data.species, species{sidx});
    calls = map(species{sidx});
    if isempty(calls) || ~callP
        if ~isempty(calls)
            % User wanted calls checked, but none exist
            callwarn = true;
        end
        retain = retain | smatch;
    else
        % further refine by call type
        cmatch = false(sum(smatch), 1);
        for cidx = 1:length(calls)
            cmatch = cmatch | strcmp(data.call(smatch), calls{cidx});
        end
        matches = smatch;
        matches(smatch) = cmatch;
        retain = retain | matches;
    end
end

keepN = sum(retain);
if keepN ~= data.N
    % something to be retained or removed
    discard = ~ retain;
    names = fieldnames(data);
    for idx = 1:length(names)
        if length(data.(names{idx})) == data.N
            % vector that needs to be adjusted
            data.(names{idx})(discard) = [];
        end
    end
    data.N = length(data.start);
end

